@@ -59,14 +59,14 @@
  *                           putawayID    4  home directory id
  */
 
-const u_char ufinderi[ADEDLEN_FINDERI] = {
+const uint8_t ufinderi[ADEDLEN_FINDERI] = {
                               0, 0, 0, 0, 0, 0, 0, 0,
                               1, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0
                           };
 
-static const u_char old_ufinderi[] = {
+static const uint8_t old_ufinderi[] = {
                               'T', 'E', 'X', 'T', 'U', 'N', 'I', 'X'
                           };
 
@@ -105,6 +105,18 @@ void *get_finderinfo(const struct vol *vol, const char *upath, struct adouble *a
             memcpy((char *)data + FINDERINFO_FRFLAGOFF, &ashort, sizeof(ashort));
         }
     }
+#ifdef __APPLE__
+    /* If no adouble finder information was found, check for native macOS info */
+    if (chk_ext && (vol->v_adouble == AD_VERSION_EA))
+    {
+        char NativeFinderInfo[ADEDLEN_FINDERI] ={'\0'};
+        if (ad_open_native_finderinfo(upath, NativeFinderInfo))
+        {
+            memcpy(data, NativeFinderInfo, ADEDLEN_FINDERI);
+            chk_ext = 0;
+        }
+    }
+#endif
 
     if (islink && !vol_syml_opt(vol)) {
         uint16_t linkflag;
@@ -251,7 +263,7 @@ uint32_t get_id(struct vol *vol,
             }
         }
         else if (adp && adcnid && (adcnid != dbcnid)) { /* (3) */
-            /* Update the ressource fork. For a folder adp is always null */
+            /* Update the resource fork. For a folder adp is always null */
             LOG(log_debug, logtype_afpd, "get_id(%s/%s): calling ad_setid(old: %u, new: %u)",
                 getcwdpath(), upath, htonl(adcnid), htonl(dbcnid));
             if (ad_setid(adp, st->st_dev, st->st_ino, dbcnid, did, vol->v_stamp)) {
@@ -279,7 +291,7 @@ int getmetadata(const AFPObj *obj,
     uint32_t		aint;
     cnid_t              id = 0;
     uint16_t		ashort;
-    u_char              achar, fdType[4];
+    uint8_t              achar, fdType[4];
     uint32_t           utf8 = 0;
     struct stat         *st;
     struct maccess	ma;
@@ -291,7 +303,7 @@ int getmetadata(const AFPObj *obj,
     data = buf;
 
     if ( ((bitmap & ( (1 << FILPBIT_FINFO)|(1 << FILPBIT_LNAME)|(1 <<FILPBIT_PDINFO) ) ) && !path->m_name)
-         || (bitmap & ( (1 << FILPBIT_LNAME) ) && utf8_encoding(obj)) /* FIXME should be m_name utf8 filename */
+         || (bitmap & (1 << FILPBIT_LNAME) && utf8_encoding(obj)) /* FIXME should be m_name utf8 filename */
          || (bitmap & (1 << FILPBIT_FNUM))) {
         if (!path->id) {
             bstring fullpath;
@@ -393,7 +405,7 @@ int getmetadata(const AFPObj *obj,
 
         case FILPBIT_MDATE :
             if ( adp && (ad_getdate(adp, AD_DATE_MODIFY, &aint) == 0)) {
-                if ((st->st_mtime > AD_DATE_TO_UNIX(aint))) {
+                if (st->st_mtime > AD_DATE_TO_UNIX(aint)) {
                    aint = AD_DATE_FROM_UNIX(st->st_mtime);
                 }
             } else {
@@ -582,7 +594,7 @@ int getmetadata(const AFPObj *obj,
             break;
 
         default :
-            return( AFPERR_BITMAP );
+            return AFPERR_BITMAP;
         }
         bitmap = bitmap>>1;
         bit++;
@@ -598,7 +610,7 @@ int getmetadata(const AFPObj *obj,
         data = set_name(vol, data, dir->d_did, path->m_name, id, utf8);
     }
     *buflen = data - buf;
-    return (AFP_OK);
+    return AFP_OK;
 }
 
 /* ----------------------- */
@@ -607,8 +619,8 @@ int getfilparams(const AFPObj *obj, struct vol *vol, uint16_t bitmap, struct pat
 {
     struct adouble	ad, *adp;
     int                 opened = 0;
-    int rc;
-    int flags; /* uninitialized ok */
+    int rc = AFP_OK;
+    int flags = 0;
 
     LOG(log_debug, logtype_afpd, "getfilparams(\"%s\")", path->u_name);
 
@@ -648,7 +660,7 @@ int getfilparams(const AFPObj *obj, struct vol *vol, uint16_t bitmap, struct pat
     if (opened)
         ad_close(adp, ADFLAGS_HF | flags);
 
-    return( rc );
+    return rc;
 }
 
 /* ----------------------------- */
@@ -671,7 +683,7 @@ int afp_createfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_,
     ibuf += sizeof( vid );
 
     if (NULL == ( vol = getvolbyvid( vid )) )
-        return( AFPERR_PARAM );
+        return AFPERR_PARAM;
 
     if (vol->v_flags & AFPVOL_RO)
         return AFPERR_VLOCK;
@@ -685,7 +697,7 @@ int afp_createfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_,
     if (NULL == ( s_path = cname( vol, dir, &ibuf )) )
         return get_afp_errno(AFPERR_PARAM);
     if ( *s_path->m_name == '\0' )
-        return( AFPERR_BADTYPE );
+        return AFPERR_BADTYPE;
 
     upath = s_path->u_name;
     ad_init(&ad, vol);
@@ -710,17 +722,17 @@ int afp_createfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_,
         case EROFS:
             return AFPERR_VLOCK;
         case ENOENT : /* we were already in 'did folder' so chdir() didn't fail */
-            return ( AFPERR_NOOBJ );
+            return AFPERR_NOOBJ;
         case EEXIST :
-            return( AFPERR_EXIST );
+            return AFPERR_EXIST;
         case EACCES :
-            return( AFPERR_ACCESS );
+            return AFPERR_ACCESS;
         case EDQUOT:
         case ENOSPC :
 	    LOG(log_info, logtype_afpd, "afp_createfile: DISK FULL");
-            return( AFPERR_DFULL );
+            return AFPERR_DFULL;
         default :
-            return( AFPERR_PARAM );
+            return AFPERR_PARAM;
         }
     }
     if ( ad_meta_fileno( &ad ) == -1 ) { /* Hard META / HF */
@@ -758,7 +770,7 @@ createfile_iderr:
     curdir->d_offcnt++;
     setvoltime(obj, vol );
 
-    return (retvalue);
+    return retvalue;
 }
 
 int afp_setfilparams(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, size_t *rbuflen)
@@ -775,7 +787,7 @@ int afp_setfilparams(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_
     memcpy(&vid, ibuf, sizeof( vid ));
     ibuf += sizeof( vid );
     if (NULL == ( vol = getvolbyvid( vid )) ) {
-        return( AFPERR_PARAM );
+        return AFPERR_PARAM;
     }
 
     if (vol->v_flags & AFPVOL_RO)
@@ -796,11 +808,11 @@ int afp_setfilparams(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_
     }
 
     if (path_isadir(s_path)) {
-        return( AFPERR_BADTYPE ); /* it's a directory */
+        return AFPERR_BADTYPE; /* it's a directory */
     }
 
     if ( s_path->st_errno != 0 ) {
-        return( AFPERR_NOOBJ );
+        return AFPERR_NOOBJ;
     }
 
     if ((intptr_t)ibuf & 1 ) {
@@ -811,7 +823,7 @@ int afp_setfilparams(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_
         setvoltime(obj, vol );
     }
 
-    return( rc );
+    return rc;
 }
 
 /*
@@ -828,8 +840,8 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
     int			bit, isad = 1, err = AFP_OK;
     char                *upath;
     char		*ade = NULL;
-    u_char              achar, xyy[4];
-    const u_char        *fdType = NULL;
+    uint8_t              achar, xyy[4];
+    const uint8_t        *fdType = NULL;
     uint16_t		ashort = 0;
     uint16_t		bshort;
     uint16_t		oshort;
@@ -845,15 +857,13 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
     gid_t		f_gid;
     uint16_t           bitmap = f_bitmap;
     uint32_t           cdate,bdate;
-    u_char              finder_buf[32];
+    uint8_t              finder_buf[32];
     int symlinked = S_ISLNK(path->st.st_mode);
     int fp;
     ssize_t len;
     char symbuf[MAXPATHLEN+1];
 
-#ifdef DEBUG
     LOG(log_debug9, logtype_afpd, "begin setfilparams:");
-#endif /* DEBUG */
 
     adp = of_ad(vol, path, &ad);
     upath = path->u_name;
@@ -871,7 +881,9 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
         }
         switch(  bit ) {
         case FILPBIT_ATTR :
-            change_mdate = 1;
+            /* Preserve file dates during a GS/OS folder copy */
+            if (obj->afp_version >= 30)
+                change_mdate = 1;
             memcpy(&ashort, buf, sizeof( ashort ));
             buf += sizeof( ashort );
             break;
@@ -941,7 +953,7 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
             memcpy( &upriv, buf, sizeof( upriv ));
             buf += sizeof( upriv );
             upriv = ntohl (upriv);
-            if ((upriv & S_IWUSR)) {
+            if (upriv & S_IWUSR) {
             	setfilunixmode(vol, path, upriv);
             }
             else {
@@ -955,10 +967,10 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
                 buf += 2;
                 /* Keep special case to support crlf translations */
                 if ((unsigned int) achar == 0x04) {
-	       	    fdType = (u_char *)"TEXT";
+	       	    fdType = (uint8_t *)"TEXT";
 		    buf += 2;
                 } else {
-            	    xyy[0] = ( u_char ) 'p';
+            	    xyy[0] = ( uint8_t ) 'p';
             	    xyy[1] = achar;
             	    xyy[3] = *buf++;
             	    xyy[2] = *buf++;
@@ -993,7 +1005,7 @@ int setfilparams(const AFPObj *obj, struct vol *vol,
         }
         LOG(log_debug, logtype_afpd, "setfilparams: no adouble perms, but only FILPBIT_MDATE and/or FILPBIT_UNIXPR");
         isad = 0;
-    } else if ((ad_get_MD_flags( adp ) & O_CREAT) ) {
+    } else if (ad_get_MD_flags( adp ) & O_CREAT ) {
         ad_setname(adp, path->m_name);
         cnid_t id;
         if ((id = get_id(vol, adp, &path->st, curdir->d_did, upath, strlen(upath))) == CNID_INVALID) {
@@ -1111,9 +1123,7 @@ setfilparam_done:
         setdirparams(vol, &Cur_Path, bitmap, (char *)&newdate);
     }
 
-#ifdef DEBUG
     LOG(log_debug9, logtype_afpd, "end setfilparams:");
-#endif /* DEBUG */
     return err;
 }
 
@@ -1139,10 +1149,10 @@ int renamefile(struct vol *vol, struct dir *ddir, int sdir_fd, char *src, char *
     if ( unix_rename( sdir_fd, src, -1, dst ) < 0 ) {
         switch ( errno ) {
         case ENOENT :
-            return( AFPERR_NOOBJ );
+            return AFPERR_NOOBJ;
         case EPERM:
         case EACCES :
-            return( AFPERR_ACCESS );
+            return AFPERR_ACCESS;
         case EROFS:
             return AFPERR_VLOCK;
         case EXDEV :			/* Cross device move -- try copy */
@@ -1156,11 +1166,11 @@ int renamefile(struct vol *vol, struct dir *ddir, int sdir_fd, char *src, char *
     	    }
             if (AFP_OK != ( rc = copyfile(vol, vol, ddir, sdir_fd, src, dst, newname, NULL )) ) {
                 /* on error copyfile delete dest */
-                return( rc );
+                return rc;
             }
             return deletefile(vol, sdir_fd, src, 0);
         default :
-            return( AFPERR_PARAM );
+            return AFPERR_PARAM;
         }
     }
 
@@ -1188,14 +1198,14 @@ int renamefile(struct vol *vol, struct dir *ddir, int sdir_fd, char *src, char *
         }
     }
 
-    /* don't care if we can't open the newly renamed ressource fork */
+    /* don't care if we can't open the newly renamed resource fork */
     if (ad_open(adp, dst, ADFLAGS_HF | ADFLAGS_RDWR) == 0) {
         ad_setname(adp, newname);
         ad_flush( adp );
         ad_close( adp, ADFLAGS_HF );
     }
 
-    return( AFP_OK );
+    return AFP_OK;
 }
 
 /* ----------------
@@ -1273,10 +1283,13 @@ int afp_copyfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, si
 {
     struct vol	*s_vol, *d_vol;
     struct dir	*dir;
-    char	*newname, *p, *upath;
+    char	*newname;
+    char	*p;
+    char	*upath = NULL;
     struct path *s_path;
     uint32_t	sdid, ddid;
-    int         err, retvalue = AFP_OK;
+    int         err = AFP_OK;
+    int         retvalue = AFP_OK;
     uint16_t	svid, dvid;
 
     struct adouble ad, *adp;
@@ -1288,7 +1301,7 @@ int afp_copyfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, si
     memcpy(&svid, ibuf, sizeof( svid ));
     ibuf += sizeof( svid );
     if (NULL == ( s_vol = getvolbyvid( svid )) ) {
-        return( AFPERR_PARAM );
+        return AFPERR_PARAM;
     }
 
     memcpy(&sdid, ibuf, sizeof( sdid ));
@@ -1306,7 +1319,7 @@ int afp_copyfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, si
         return get_afp_errno(AFPERR_PARAM);
     }
     if ( path_isadir(s_path) ) {
-        return( AFPERR_BADTYPE );
+        return AFPERR_BADTYPE;
     }
 
     /* don't allow copies when the file is open.
@@ -1397,8 +1410,11 @@ int afp_copyfile(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U_, si
     setvoltime(obj, d_vol );
 
 copy_exit:
+    if (upath && err == AFP_OK) {
+        fce_register(obj, FCE_FILE_CREATE, fullpathname(upath), NULL);
+    }
     ad_close( adp, ADFLAGS_DF |ADFLAGS_HF | ADFLAGS_SETSHRMD);
-    return( retvalue );
+    return retvalue;
 }
 
 /* ----------------------------------
@@ -1461,10 +1477,18 @@ int copyfile(struct vol *s_vol,
         }
         return AFPERR_EXIST;
     }
+    size_t copybuf_len = 0;
+    uint8_t* copybuf = NULL;
+
+    if (s_vol->v_obj->proto == AFPPROTO_DSI)
+    {
+        copybuf = s_vol->v_obj->dsi->commands;
+        copybuf_len = s_vol->v_obj->dsi->server_quantum;
+    }
 
     if (copy_fork(ADEID_DFORK, &add, adp,
-                         s_vol->v_obj->dsi->commands,
-                         s_vol->v_obj->dsi->server_quantum) != 0) {
+                         copybuf,
+                         copybuf_len) != 0) {
         err = errno;
         LOG(log_error, logtype_afpd, "copyfile('%s'): copy_fork: %s", src, strerror(errno));
     } else if (d_vol->vfs->vfs_copyfile(d_vol, sfd, src, dst) != AFP_OK) {
@@ -1553,7 +1577,7 @@ uint16_t   bshort = 0;
 	if (!(vol->v_ignattr & ATTRBIT_NODELETE) && (bshort & htons(ATTRBIT_NODELETE))) {
 		return AFPERR_OLOCK;
 	}
-     if ((bshort & htons(ATTRBIT_DOPEN | ATTRBIT_ROPEN))) {
+     if (bshort & htons(ATTRBIT_DOPEN | ATTRBIT_ROPEN)) {
     	return AFPERR_BUSY;
 	}
 	return 0;
@@ -1610,10 +1634,10 @@ int deletefile(const struct vol *vol, int dirfd, char *file, int checkAttrib)
     if ( adp && AD_RSRC_OPEN(adp) ) { /* there's a resource fork */
         adflags |= ADFLAGS_RF;
         /* FIXME we have a pb here because we want to know if a file is open
-         * there's a 'priority inversion' if you can't open the ressource fork RW
+         * there's a 'priority inversion' if you can't open the resource fork RW
          * you can delete it if it's open because you can't get a write lock.
          *
-         * ADLOCK_FILELOCK means the whole ressource fork, not only after the
+         * ADLOCK_FILELOCK means the whole resource fork, not only after the
          * metadatas
          *
          * FIXME it doesn't work for RFORK open read only and fork open without deny mode
@@ -1672,7 +1696,7 @@ int afp_createid(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf, si
     ibuf += sizeof(vid);
 
     if (NULL == ( vol = getvolbyvid( vid )) ) {
-        return( AFPERR_PARAM);
+        return AFPERR_PARAM;
     }
 
     if (vol->v_cdb == NULL || !(vol->v_cdb->cnid_db_flags & CNID_FLAG_PERSISTENT)) {
@@ -1694,7 +1718,7 @@ int afp_createid(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf, si
     }
 
     if ( path_isadir(s_path) ) {
-        return( AFPERR_BADTYPE );
+        return AFPERR_BADTYPE;
     }
 
     upath = s_path->u_name;
@@ -1811,7 +1835,7 @@ int afp_resolveid(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf, size_
     ibuf += sizeof(vid);
 
     if (NULL == ( vol = getvolbyvid( vid )) ) {
-        return( AFPERR_PARAM);
+        return AFPERR_PARAM;
     }
 
     if (vol->v_cdb == NULL || !(vol->v_cdb->cnid_db_flags & CNID_FLAG_PERSISTENT)) {
@@ -1919,7 +1943,7 @@ int afp_deleteid(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _U_
     ibuf += sizeof(vid);
 
     if (NULL == ( vol = getvolbyvid( vid )) ) {
-        return( AFPERR_PARAM);
+        return AFPERR_PARAM;
     }
 
     if (vol->v_cdb == NULL || !(vol->v_cdb->cnid_db_flags & CNID_FLAG_PERSISTENT)) {
@@ -1945,7 +1969,7 @@ int afp_deleteid(AFPObj *obj _U_, char *ibuf, size_t ibuflen _U_, char *rbuf _U_
             err = AFPERR_NOOBJ;
             goto delete;
         }
-        return( AFPERR_PARAM );
+        return AFPERR_PARAM;
     }
 
     err = AFP_OK;
@@ -2065,10 +2089,10 @@ int afp_exchangefiles(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U
     ibuf += sizeof(vid);
 
     if (NULL == ( vol = getvolbyvid( vid )) ) {
-        return( AFPERR_PARAM);
+        return AFPERR_PARAM;
     }
 
-    if ((vol->v_flags & AFPVOL_RO))
+    if (vol->v_flags & AFPVOL_RO)
         return AFPERR_VLOCK;
 
     /* source and destination dids */
@@ -2226,7 +2250,8 @@ int afp_exchangefiles(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U
 
     if (AD_META_OPEN(adsp) || AD_META_OPEN(addp)) {
         struct adouble adtmp;
-        bool opened_ads, opened_add;
+        bool opened_ads = false;
+        bool opened_add = false;
 
         ad_init(&adtmp, vol);
         ad_init_offsets(&adtmp);
@@ -2258,7 +2283,7 @@ int afp_exchangefiles(AFPObj *obj, char *ibuf, size_t ibuflen _U_, char *rbuf _U
             ad_close(addp, ADFLAGS_HF);
     }
 
-    /* FIXME: we should switch ressource fork too */
+    /* FIXME: we should switch resource fork too */
 
     /* here we need to reopen if crossdev */
     if (sid && ad_setid(addp, destst.st_dev, destst.st_ino,  sid, sdir->d_did, vol->v_stamp))
