@@ -28,17 +28,17 @@ int route( int message, struct sockaddr *dst, struct sockaddr *gate, int flags)
     rtent.rt_dst = *dst;
     rtent.rt_gateway = *gate;
     rtent.rt_flags = flags;
-    return( ioctl( rtfd, message, &rtent ));
+    return ioctl( rtfd, message, &rtent );
 }
 
 #else /* BSD4_4 */
 
 struct sockaddr_m {
-    u_char	sam_len;
-    u_char	sam_family;
-    u_short	sam_pad;
-    u_short	sam_mask;
-    u_short	sam_pad2;
+    unsigned char	sam_len;
+    unsigned char	sam_family;
+    unsigned short	sam_pad;
+    unsigned short	sam_mask;
+    unsigned short	sam_pad2;
 } mask = { sizeof( struct sockaddr_m ), 0, 0, 0xffff, 0 };
 
 struct rt_msg_at {
@@ -72,8 +72,8 @@ int route( int message, struct sockaddr_at	*dst, struct sockaddr_at *gate, int f
     rtma.rtma_gate = *gate;
     if (( rc = write( rtfd, &rtma, rtma.rtma_rtm.rtm_msglen )) !=
 	    rtma.rtma_rtm.rtm_msglen ) {
-	return( -1 );
+	return -1;
     }
-    return( 0 );
+    return 0;
 }
 #endif /* BSD4_4 */
